;
; Program flow macros
;

TOGGLE_TRAP		EQU	13
BREAK_COOPERATE_TRAP	EQU	14

TOGGLE_PROSIT	MACRO
		trap		#TOGGLE_TRAP
		ENDM

BREAK_AND_COOPERATE	MACRO
		trap		#BREAK_COOPERATE_TRAP
		ENDM

TRIGGER_EVENT	MACRO	; processID, event
		move.l		\2,-(sp)
		move.w		\1,-(sp)
		jsr		Prosit::Trigger_Event
		addq.l		#6,sp
		ENDM

BROADCAST_EVENT	MACRO	; event
		move.l		\1,-(sp)
		jsr		Prosit::Broadcast_Event
		addq.l		#4,sp
		ENDM



;
; Prosit header macros
;

; First byte = size of field, second byte = field ID.
; Reserved values: $xx00 and $xxFF
PROSIT_VERSION		EQU	$0401
PROSIT_FILETYPE		EQU	$0402
PROSIT_STACKSIZE	EQU	$0403
PROSIT_EVENT		EQU	$0A10

PROSIT_VERSION_STRING		EQU	"0702"
PROSIT_FILETYPE_PROGRAM		EQU	"prog"
PROSIT_FILETYPE_BACKGROUND	EQU	"bkgr"

HEADER_BEGIN MACRO
	dc.w	$02FF
	ENDM

DEFINE_PROSIT_VERSION MACRO
	dc.w	PROSIT_VERSION
	dc.l	\1
	ENDM

DEFINE_PROSIT_FILETYPE MACRO
	dc.w	PROSIT_FILETYPE
	dc.l	\1
	ENDM

DEFINE_PROSIT_STACKSIZE MACRO
	dc.w	PROSIT_STACKSIZE,\1
	ENDM

DEFINE_PROSIT_EVENT MACRO
	dc.w	PROSIT_EVENT
	dc.l	\1,\2
	ENDM

HEADER_END MACRO
	dc.w	$0000
	ENDM



;
; Lock macros
;

PROSIT_LOCK_HEADER	MACRO	; Lockvar #
	EVEN
	P_Lock_\1:
		add.b		#1,p_lock_var_\1
		beq.s		P_Lock_Run_\1
		sub.b		#1,p_lock_var_\1
		BREAK_AND_COOPERATE
		bra.s		P_Lock_\1
	P_Lock_Run_\1:
		rts
	P_Unlock_\1:
		sub.b		#1,p_lock_var_\1
		rts
	p_lock_var_\1:
		dc.b	-1
	EVEN
		ENDM

PROSIT_LOCK	MACRO	; Lockvar #
		bsr		P_Lock_\1
		ENDM

PROSIT_UNLOCK	MACRO	; Lockvar #
		bsr		P_Unlock_\1
		ENDM


;*********************************************************************
;*                         SYSTEM FUNCTIONS                          *
;*********************************************************************

;---------------------------------------------------------------------
; BYTE *Header_Get(WORD filehandle)
;
; Function: Get the header of a Prosit file.
;
; Output: a0 = *header, 0 if none found.
;---------------------------------------------------------------------
Prosit::Header_Get		equ	prosit@0000

;---------------------------------------------------------------------
; BYTE *Header_FindField(BYTE *current, WORD type)
;
; Function: Find the next header field of <type> in a Prosit header.
;  Call with a pointer to the header to find the first field.
;  Call repeatedly with the return value to find subsequent fields.
;
; Output: a0 = *field, 0 if none found.
;---------------------------------------------------------------------
Prosit::Header_FindField	equ	prosit@0001

;---------------------------------------------------------------------
; Process_New(WORD filehandle)
;
; Function: Create a new process from the file with a specified handle
;
; Output: d0 = Process ID, -1 if failure
;---------------------------------------------------------------------
Prosit::Process_New		equ	prosit@0002

;---------------------------------------------------------------------
; Process_Fork(BYTE *header, BYTE *code)
;
; Function: Fork an existing process
;
; Output: d0 = Process ID, -1 if failure
;---------------------------------------------------------------------
Prosit::Process_Fork		equ	prosit@0003

;---------------------------------------------------------------------
; Process_Terminate(WORD ID)
;
; Function: Terminate process
;
; Output: d0 = 0 if success, -1 if failure
;---------------------------------------------------------------------
Prosit::Process_Terminate	equ	prosit@0004

;---------------------------------------------------------------------
; Sleep(void)
;
; Function: Put process to sleep, i.e. move it from the list of
;  executed processes to the list of sleeping processes. It will
;  automatically wake up when a relevant event is triggered.
;---------------------------------------------------------------------
Prosit::Sleep			equ	prosit@0005

;---------------------------------------------------------------------
; Wake_Up(WORD ID)
;
; Function: Wake up process <ID>, i.e. move it from the list of
;  sleeping processes to the list of executed processes. If the
;  process isn't sleeping nothing happens.
;---------------------------------------------------------------------
Prosit::Wake_Up			equ	prosit@0006

;---------------------------------------------------------------------
; Trigger_Event(WORD ID, LONG event)
;
; Function: Execute the <event> subroutine inside process <ID>.
;
; Output: d0.w = 0 if success, -1 if failed
;---------------------------------------------------------------------
Prosit::Trigger_Event		equ	prosit@0007

;---------------------------------------------------------------------
; Broadcast_Event(LONG event)
;
; Function: Execute the <event> subroutine inside all processes
;  except the caller.
;---------------------------------------------------------------------
Prosit::Broadcast_Event		equ	prosit@0008

;---------------------------------------------------------------------
; Get_File_Handle(BYTE *foldername, BYTE *filename)
;
; Function: Get the handle of a specific file
;
; Output: d5 = handle, 0 if file doesn't exist
;---------------------------------------------------------------------
Prosit::Get_File_Handle		equ	prosit@0009

;---------------------------------------------------------------------
; Execute_File(WORD handle)
;
; Function: Execute a file.
;
; Output: d0.w = 0 if success, -1 if failed
;---------------------------------------------------------------------
Prosit::Execute_File		equ	prosit@000A

;*********************************************************************
;*                         SYSTEM VARIABLES                          *
;*********************************************************************

;---------------------------------------------------------------------
;WORD: CurrentProcessID
;Current process' ID (for subroutines and others)
;---------------------------------------------------------------------
Prosit::CurrentProcessID	equ	prosit@000B

;---------------------------------------------------------------------
;LONG: Ticks
;Number of ticks since Prosit was started. Is contained in a longword
;so it restarts after a while, but by then the batteries will probably
;be dead :)
;---------------------------------------------------------------------
Prosit::Ticks			equ	prosit@000C

;---------------------------------------------------------------------
;BYTE: ControlKeys
;The mask of the row in the keyboard matrix that contains the arrow-
;and shiftkeys:
; Bit 0: 2nd
; Bit 1: Diamond
; Bit 2: Shift
; Bit 3: Hand
; Bit 4: Left
; Bit 5: Up
; Bit 6: Right
; Bit 7: Down
;---------------------------------------------------------------------
Prosit::ControlKeys		equ	prosit@000D

;---------------------------------------------------------------------
;WORD: CurrentKey
;Holds the keycode of the current key being held down, excluding the
;arrow keys.
;---------------------------------------------------------------------
Prosit::CurrentKey		equ	prosit@000E




;*********************************************************************
;*                     LIST HEADERS AND FUNCTIONS                    *
;*********************************************************************
LIST_NEXT		EQU	0
LIST_ID			EQU	LIST_NEXT+4
LIST_PRIORITY		EQU	LIST_ID+2
LIST_BLANK		EQU	LIST_PRIORITY+1
LIST_HEADER_SIZE	EQU	LIST_BLANK+1

;---------------------------------------------------------------------
; List_Insert(BYTE *list, BYTE *structure, BYTE priority)
;
; Function: Insert <structure> in <list>. Its placement depends on
;  <priority>.
;---------------------------------------------------------------------
Prosit::List_Insert		equ	prosit@000F

;---------------------------------------------------------------------
; List_Remove(BYTE *list, BYTE *structure)
;
; Function: Remove <structure> from <list>.
;---------------------------------------------------------------------
Prosit::List_Remove		equ	prosit@0010

;---------------------------------------------------------------------
; List_Previous(BYTE *structure)
;
; Function: Get a pointer to the structure in <list> that points to
; <structure>
;
; Output: a0.l = *previous structure
;---------------------------------------------------------------------
Prosit::List_Previous		equ	prosit@0011

;---------------------------------------------------------------------
; List_Find(BYTE *list, WORD ID)
;
; Function: Get a pointer to the structure in <list> with <ID>.
;
; Output: a0.l = *structure, 0 if not found
;---------------------------------------------------------------------
Prosit::List_Find		equ	prosit@0012

;---------------------------------------------------------------------
; List_Size(BYTE *list)
;
; Function: Get number of elements in list.
;
; Output: d0.w = No. of elements in list.
;---------------------------------------------------------------------
Prosit::List_Size		equ	prosit@0013
