#ifndef prosit_h
#define prosit_h

#define PROSIT_VERSION_STRING "0800"
#define PROSIT_VERSION_STRING_N 0x30383030	// "0800"

#include "romcalls.h"

#ifdef TI92PLUS
  typedef union {
    WORD data;
    struct {
        unsigned int down : 1;
        unsigned int right : 1;
        unsigned int up : 1;
        unsigned int left : 1;
        unsigned int hand : 1;
        unsigned int shift : 1;
        unsigned int diamond : 1;
        unsigned int second : 1;

        unsigned int flag0_7 : 1;
        unsigned int flag0_6 : 1;
        unsigned int flag0_5 : 1;
        unsigned int flag0_4 : 1;
        unsigned int flag0_3 : 1;
        unsigned int flag0_2 : 1;
        unsigned int flag0_1 : 1;
        unsigned int flag0_0 : 1;
    } flags;
  } ControlKeysStruct;
  int _ti92plus;
#endif

#ifdef TI89
  typedef union {
    WORD data;
    struct {
        unsigned int alpha : 1;
        unsigned int diamond : 1;
        unsigned int shift : 1;
        unsigned int second : 1;
        unsigned int right : 1;
        unsigned int down : 1;
        unsigned int left : 1;
        unsigned int up : 1;

        unsigned int flag0_7 : 1;
        unsigned int flag0_6 : 1;
        unsigned int flag0_5 : 1;
        unsigned int flag0_4 : 1;
        unsigned int flag0_3 : 1;
        unsigned int flag0_2 : 1;
        unsigned int flag0_1 : 1;
        unsigned int flag0_0 : 1;
    } flags;
  } ControlKeysStruct;
  int _ti89;
#endif


#define STR(x) #x
#define XSTR(x) STR(x)
#define S2L(str) ((long)str[0] << 24) + ((long)str[1] << 16) + ((long)str[2] << 8) + ((long)str[3])
#define S2W(str) ((long)str[0] << 8) + ((long)str[1])

typedef struct {
  BYTE data;
} MEMORY_BYTE;

typedef struct {
  WORD data;
} MEMORY_WORD;

typedef struct {
  ULONG data;
} MEMORY_LONG;


//
// Prosit executable file structure
//
typedef struct {
  WORD filesize;
  WORD rts;
  WORD offset_prosit;
  ULONG signature;
  WORD reloccount;
  WORD offset_comment;
  WORD offset_main;
  WORD offset_exit;
  union {
    WORD all;
    struct {
      unsigned int ti89_emulation : 1;
      unsigned int flag1_6 : 1;
      unsigned int flag1_5 : 1;
      unsigned int flag1_4 : 1;
      unsigned int flag1_3 : 1;
      unsigned int flag1_2 : 1;
      unsigned int flag1_1 : 1;
      unsigned int copy_of_archived : 1;

      unsigned int flag0_7 : 1;
      unsigned int flag0_6 : 1;
      unsigned int flag0_5 : 1;
      unsigned int flag0_4 : 1;
      unsigned int flag0_3 : 1;
      unsigned int flag0_2 : 1;
      unsigned int runs_on_92plus : 1;
      unsigned int runs_on_89 : 1;
    } bits;
  } flags;
} PROSIT_FILE_HEADER;

typedef struct ListElementHeaderStruct {
  struct ListElementHeaderStruct *next;
  WORD id;
  signed char priority;
  BYTE flags;
} ListElementHeader;

typedef struct ListHeaderStruct {
  ListElementHeader *start;
} ListHeader;

//
// Prosit header
//
#define PROSIT_FILETYPE_PROGRAM "prog"
#define PROSIT_FILETYPE_PROGRAM_N 0x70726f67	// "prog"
#define PROSIT_FILETYPE_BACKGROUND "bkgr"
#define PROSIT_FILETYPE_BACKGROUND_N 0x626b6772	// "bkgr"

#define PROSIT_BEGIN_HEADER 0x02ff
#define PROSIT_STACKSIZE 0x0401
#define PROSIT_VERSION 0x0601
#define PROSIT_FILETYPE 0x0602
#define PROSIT_EVENT 0x0a01
#define PROSIT_END_HEADER 0x0000

#define BEGIN_PROSIT_HEADER(name) asm(" .even"); char name[0] = ""; asm("\n .word " XSTR(PROSIT_BEGIN_HEADER) "\n"
#define DEFINE_PROSIT_STACKSIZE(size) " .word " XSTR(PROSIT_STACKSIZE) ", " XSTR(size) "\n"
#define DEFINE_PROSIT_VERSION(version) " .word " XSTR(PROSIT_VERSION) "\n .ascii \"" version "\"\n"
#define DEFINE_PROSIT_FILETYPE(type) " .word " XSTR(PROSIT_FILETYPE) "\n .ascii \"" type "\"\n"
#define DEFINE_PROSIT_EVENT(name,function) ".word " XSTR(PROSIT_EVENT) "\n .long " XSTR(name) "\n .long " #function "\n"
#define END_PROSIT_HEADER " .word " XSTR(PROSIT_END_HEADER));

//
// Event types
//
#define QUIT_EVENT 0x71756974		// "quit"
#define MOUSEBUTTON_EVENT 0x6d6f7562	// "moub"

typedef union {
  WORD fieldtype;
  struct {
    BYTE next;
    BYTE type;
  } parts;
} PrositHeaderElement_Header;

typedef struct {
  PrositHeaderElement_Header header;
  WORD stacksize;
} PrositHeaderElement_Stacksize;

typedef struct {
  PrositHeaderElement_Header header;
  ULONG version;
} PrositHeaderElement_Version;

typedef struct {
  PrositHeaderElement_Header header;
  ULONG filetype;
} PrositHeaderElement_Filetype;

typedef struct {
  PrositHeaderElement_Header header;
  ULONG event;
  void *code;
} PrositHeaderElement_Event;

//
// Prosit process context structure
//
typedef struct PROCESS_CONTEXT_ENTRY_STRUCT {
  ListElementHeader header;
  ULONG registers[15];
  void *usp;
  WORD sr;
  void *pc;
  ULONG tiosdata[9];
  union {
    WORD all;
    struct {
      unsigned int flag1_7 : 1;
      unsigned int flag1_6 : 1;
      unsigned int flag1_5 : 1;
      unsigned int flag1_4 : 1;
      unsigned int flag1_3 : 1;
      unsigned int flag1_2 : 1;
      unsigned int flag1_1 : 1;
      unsigned int flag1_0 : 1;

      unsigned int flag0_7 : 1;
      unsigned int flag0_6 : 1;
      unsigned int flag0_5 : 1;
      unsigned int flag0_4 : 1;
      unsigned int fork : 1;
      unsigned int copy_of_archived : 1;
      unsigned int in_interrupt : 1;
      unsigned int sleeping : 1;
    } bits;
  } flags;
  HANDLE filehandle;
  HANDLE stackhandle;
  void *prositheader;
} PROCESS_CONTEXT_ENTRY;

#define CS_INIT_NORMAL		0x0000
#define CS_INIT_FORK		0x0008
#define CS_FORK_KEEP_FLAGS	0x0004

#define ToggleProsit() asm(" trap #13");
#define BreakAndCooperate() asm(" trap #14");







//
// Functions exported by Prosit
//
#define CurrentProcessID prosit__0000
extern HANDLE CurrentProcessID;

#define Ticks prosit__0001
extern volatile ULONG Ticks;

#define CurrentKey prosit__0002
extern volatile WORD CurrentKey;

#define ControlKeys prosit__0003
extern volatile ControlKeysStruct ControlKeys;

#define ExecuteFile prosit__0004
void executeFile(HANDLE filehandle);

#define GetPrositHeaderOfFile prosit__0005
void *GetPrositHeaderOfFile(HANDLE filehandle);

#define GetPrositHeaderOfProcess prosit__0006
void *GetPrositHeaderOfProcess(HANDLE pid);

#define FindPrositHeaderField prosit__0007
void *FindPrositHeaderField(PrositHeaderElement_Header *current, WORD type);

#define ListInitialize prosit__0008
void ListInitialize(ListHeader *List);

#define ListPrevious prosit__0009
ListElementHeader *ListPrevious(ListElementHeader *elementheader);

#define ListInsert prosit__000A
void ListInsert(ListHeader *header, ListElementHeader *elementheader);

#define ListRemove prosit__000B
void ListRemove(ListHeader *header, ListElementHeader *elementheader);

#define ListFind prosit__000C
ListElementHeader *ListFind(ListHeader *header, WORD idstring);

#define ListSize prosit__000D
int ListSize(ListHeader *header);

#define TerminateProcess prosit__000E
void TerminateProcess(HANDLE pid);

#define NewProcess prosit__000F
HANDLE NewProcess(HANDLE filehandle);

#define ForkProcess prosit__0010
HANDLE ForkProcess(void *header, void *entrypoint);

#define Sleep prosit__0011
void Sleep(HANDLE pid);

#define WakeUp prosit__0012
void WakeUp(HANDLE pid);

#define TriggerEvent prosit__0013
BOOL TriggerEvent(HANDLE target, ULONG event);

#define BroadcastEvent prosit__0014
void BroadcastEvent(ULONG event);

#endif
