#ifndef winlib_h
#define winlib_h

#include <prosit.h>

#ifdef TI92PLUS
  #define LCD_WIDTH 240
  #define LCD_HEIGHT 128
  #define ARROWKEYSMASK 0xf000
  #define MOUSEKEY ControlKeys.flags.hand
#endif

#ifdef TI89
  #define LCD_WIDTH 160
  #define LCD_HEIGHT 100
  #define ARROWKEYSMASK 0x0f00
  #define MOUSEKEY ControlKeys.flags.second
#endif

#define GOT_FOCUS_EVENT 0x77677466	// "wgtf"
#define LOST_FOCUS_EVENT 0x776c7366	// "wlsf"

enum WMReturnValues {WMR_GOTFOCUS=0x0001, WMR_MOVETOFRONT=0x0002, WMR_CLIENT=0x0004};

typedef struct {
  WORD Key;
  void *w;
  WORD ButtonID;
  int MouseX;
  int MouseY;
} PWINDOW_DATARETURN;

typedef struct PWindowStruct {
  ListElementHeader header;

  int X, Y;
  int Width, Height;
  WORD WidthInBytes;

  ListHeader StructureList;
  PWINDOW_DATARETURN *Datareturn;

  HANDLE MaskHandle, ContentsHandle[2];
  WORD VisibleRows;
  WORD VisibleColumns;
  WORD OffsetInContents;
  WORD LCDStartOffset, LCDMaskStartOffset;

  WINDOW Win;

} PWINDOW;

#define EMPTYPWINDOWSTRUCT {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, \
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0, 0, \
   0, 0, 0}


typedef ListElementHeader PWINDOW_LIST;

typedef struct {
  PWINDOW_LIST *GUI;
  PWINDOW_LIST *GotFocus;
  PWINDOW_LIST *LostFocus;
} PWINDOW_STRUCTURE;

typedef struct {
  WORD Width, Height, HotSpotX, HotSpotY;
  ICON GFX[2], Mask;
} MOUSEPOINTER;


#define B_TOP_WHITE 		0x02
#define B_TOP_BLACK 		0x03
#define B_TOP_PATTERN 		0x01
#define B_BOTTOM_WHITE 		0x08
#define B_BOTTOM_BLACK 		0x0c
#define B_BOTTOM_PATTERN 	0x04
#define B_LEFT_WHITE		0x20
#define B_LEFT_BLACK		0x30
#define B_LEFT_PATTERN		0x10
#define B_RIGHT_WHITE		0x80
#define B_RIGHT_BLACK		0xc0
#define B_RIGHT_PATTERN		0x40
#define B_FILLED_WHITE		0x02
#define B_FILLED_BLACK		0x03
#define B_FILLED_PATTERN	0x01
#define B_FRAMED_BLACK 		B_TOP_BLACK+B_BOTTOM_BLACK+B_LEFT_BLACK+B_RIGHT_BLACK
#define B_FRAMED_WHITE		B_TOP_WHITE+B_BOTTOM_WHITE+B_LEFT_WHITE+B_RIGHT_WHITE

#define R_AREA 0
#define R_TEXT_BUTTON 1
#define R_GFX_BUTTON 2
#define R_HORIZ_SLIDER 3
#define R_VERT_SLIDER 4
#define R_TEXT 64
#define R_BOX 65
#define R_SETCLIP 9

#define RF_PRESS 0x8000
#define RF_HOLD 0x4000
#define RF_RELEASE 0x2000

#define F_NORMAL	0x00
#define F_DIMMED	0x01
#define F_BOLD		0x02
#define F_UNDERLINED	0x04
#define F_INVERTED	0x08
#define F_ITALIC	0x10



typedef USHORT window;

typedef struct{BYTE x, y;
              }COORD;

typedef struct{BYTE width, height;
              }SIZE;

typedef struct{BYTE flags, font;
              }FLAGSFONT;

typedef struct{USHORT WKey;
               USHORT MouseWindow;
               USHORT MouseButton;
               USHORT MouseX;
               USHORT MouseY;
}windowreturn;

typedef struct{void *pointer;
               USHORT returnvalue;
}windowarea;

typedef struct{BYTE type;
               BYTE x, y, width, height;
              }windowobject_area;


// *********************** EXPORTED VARIABLES AND FUNCTIONS ******************************

#define WindowInFocus winlib__0000
extern PWINDOW *WindowInFocus;

#define BackgroundWindow winlib__0001
extern PWINDOW BackgroundWindow;

#define BackgroundWindowRect winlib__0002
extern WIN_RECT BackgroundWindowRect;

#define MouseX winlib__0003
extern int MouseX;

#define MouseY winlib__0004
extern int MouseY;

#define MouseInside winlib__0005
BOOL MouseInside(PWINDOW *w, SCR_RECT *r);

#define InitWindowSystem winlib__0006
BOOL InitWindowSystem(MOUSEPOINTER *mp);

#define CloseWindowSystem winlib__0007
void CloseWindowSystem(void);

#define PWinOpen winlib__0008
BOOL PWinOpen(PWINDOW *w, WIN_RECT *dimensions, PWINDOW_DATARETURN *wr, char *title);

#define PWinClose winlib__0009
void PWinClose(PWINDOW *w);

#define PWinRefresh winlib__000A
void PWinRefresh(PWINDOW *w);

#define PWinSelectPlane winlib__000B
void PWinSelectPlane(PWINDOW *w, int plane);

#define PWinMoveTo winlib__000C
void PWinMoveTo(PWINDOW *w, int newx, int newy);

#define PWinMoveToFront winlib__000D
void PWinMoveToFront(PWINDOW *w);

#endif
