#define TI92PLUS

#include <romcalls.h>
#include <prosit.h>

char _comment[] = "Prosit launcher";
char BootFileName[] = "\0proboot";

// Exported variables
HANDLE CurrentProcessID = 47;
ULONG Ticks = 0;
WORD CurrentKey = 0;
ControlKeysStruct ControlKeys = {0};

#ifdef TI92PLUS
  #define TIOS_CONSTANTS 0x00007464
  int _ti92plus;
  int KeyCodes[80] =
  {0,0,0,0,337,338,340,344,        // 2nd, Diamond, Shift, Hand, Left, Up, Right, Down
   0,112,115,119,275,49,50,51,     // N/A, z, s, w, F8, 1, 2, 3
   0,120,100,101,270,52,53,54,     // N/A, x, d, e, F3, 4, 5, 6
   258,99,102,114,274,55,56,57,    // Store, c, f, r, F7, 7, 8, 9
   32,118,103,116,269,40,41,44,    // Space, v, g, t, F2, (, ), ","
   47,98,104,121,273,259,260,261,  // "/", b, h, y, F6, Sin, Cos, Tan
   94,110,106,117,268,262,-13,112, // "^", n, j, u, F1, ln, Enter2, P
   61,109,107,105,272,263,265,42,  // "=", m, k, i, F5, Clear, Apps, "*"
   257,136,108,111,43,266,264,0,   // Backspace, theta, l, o, "+", Mode, ESC, N/A
   45,13,97,113,271,48,46,173};    // "-", Enter1, a, q, F4, 0, ".", "(-)"
#endif

#ifdef TI89
  #define TIOS_CONSTANTS 0x00007460
  int _ti89;
  int KeyCodes[80] =
  {338,337,344,340,0,0,0,0,        // Up, Left, Down, Right, 2nd, Shift, Diamond, Alpha
   13,43,45,42,47,94,263,272,      // Enter, "+", "-", "*", "/", "^", Clear, F5
   173,51,54,57,44,116,257,271,    // "(-)", 3, 6, 9, ",", t, Backspace, F4
   46,50,53,56,41,122,278,270,     // ".", 2, 5, 8, ), <, Catalog, F3
   48,49,52,55,40,121,266,269,     // 0, 1, 4, 7, (, y, Mode, F2
   265,258,149,124,61,120,277,268, // Apps ,Store, EE, "|" ,"=", x, Home, F1
   264,0,0,0,0,0,0,0,              // ESC, N/A, N/A, N/A, N/A, N/A, N/A, N/A
   0,0,0,0,0,0,0,0,                // N/A, N/A, N/A, N/A, N/A, N/A, N/A, N/A
   0,0,0,0,0,0,0,0,                // N/A, N/A, N/A, N/A, N/A, N/A, N/A, N/A
   0,0,0,0,0,0,0,0};               // N/A, N/A, N/A, N/A, N/A, N/A, N/A, N/A
#endif

// Uninitialized global variables are allocated on the BSS!
ListHeader ContextList, SleepingContextList;
WORD KeyboardDelay;

PROCESS_CONTEXT_ENTRY *CurrentProcessContext;
ULONG old_int1, old_int2, old_int3, old_int4, old_int5, old_int6, old_int7,
      old_trap13, old_trap14;

/*****************
* FILE FUNCTIONS *
*****************/
#include "prosit_file.h"

/*******************
* HEADER FUNCTIONS *
*******************/
#include "prosit_header.h"

/*****************
* LIST FUNCTIONS *
*****************/
#include "prosit_list.h"

/****************
* THE SCHEDULER *
****************/
#include "prosit_scheduler.h"

/****************
* THE KEYREADER *
****************/
#include "prosit_keyreader.h"

/*****************************
* MISC. TRAPS AND INTERRUPTS *
*****************************/

// A dummy interrupt - does nothing.
asm("	DummyInt:				\n\
		rte				");

// The "toggle system" interrupt.
asm("	ToggleProsit:				\n\
		eor.w		#0x2700,(%sp)	\n\
		rte				");

/********************
* PROCESS FUNCTIONS *
********************/
#include "prosit_process.h"

/******************
* EVENT FUNCTIONS *
******************/
#include "prosit_event.h"


/*******************
* THE MAIN PROGRAM *
*******************/
void _main(void)
{
  SYM_ENTRY *BootFileSymEntry;
  PROCESS_CONTEXT_ENTRY *BootFileContext;
  PROSIT_FILE_HEADER *BootFileStart;
  HANDLE BootFileHandle, BootFileContextHandle;

  if((BootFileSymEntry = SymFindPtr(BootFileName+8,0))) {
    if((BootFileStart = RelocateFile(BootFileHandle = BootFileSymEntry->handle))) {
      BootFileContext = (PROCESS_CONTEXT_ENTRY *)HeapDeref(CurrentProcessID = HeapAlloc(sizeof(PROCESS_CONTEXT_ENTRY)));
      CurrentProcessContext = BootFileContext;
      BootFileContext->filehandle = BootFileHandle;
      BootFileContext->header.id = CurrentProcessID;

//      CurrentProcessID = BootFileContext->header.id;

      // Setup context entry for ProBoot
      BootFileContext->header.priority = 0;
      BootFileContext->flags.all = CS_INIT_NORMAL;
      BootFileContext->stackhandle = 0;
      BootFileContext->prositheader = (void *)BootFileStart+BootFileStart->offset_prosit+2;

      // Setup the context list itself
      ListInitialize(&ContextList);
      ListInitialize(&SleepingContextList);
      ListInsert(&ContextList,&BootFileContext->header);

      // Set some variables to their default values
      CurrentKey = 0;
      ControlKeys.data = 0xff00;
      KeyboardDelay = 0;
      Ticks = 0;

      // Initialize the scheduler and the keyreader, plus set the two traps we use.
      asm("	move.w		#0x0700,%%d0
		trap		#1
		move.l		(0x00000064),old_int1
		move.l		(0x00000068),old_int2
		move.l		(0x0000006C),old_int3
		move.l		(0x00000070),old_int4
		move.l		(0x00000074),old_int5
		move.l		(0x00000078),old_int6
		move.l		(0x0000007C),old_int7
		move.l		(0x000000b4),old_trap13
		move.l		(0x000000b8),old_trap14
		bclr.b		#2,(0x00600001)
		move.l		#PreScheduler,(0x00000064)
		move.l		#DummyInt,(0x00000068)
		move.l		#DummyInt,(0x0000006C)
		move.l		#DummyInt,(0x00000070)
		move.l		#KeyReader,(0x00000074)
		move.l		#DummyInt,(0x00000078)
		move.l		#DummyInt,(0x0000007C)
		move.l		#ToggleProsit,(0x000000b4)
		move.l		#PreScheduler,(0x000000b8)
		bset.b		#2,(0x00600001)
		trap		#1
		move.b		#0xf0,(0x00600017)" : : : "d0");

      // Now we turn the control over to ProBoot.
      asm("	move.l 		%0,%%a0
		jsr		(%%a0)" : :
		"a" ((void *)BootFileStart+BootFileStart->offset_main+2) :
		"0");

      // Restore all interrupts and reset the int5 timer to the TIOS value.
      asm("	move.w		#0x0700,%%d0
		trap		#1
		bclr.b		#2,(0x00600001)
		move.l		old_int1,(0x00000064)
		move.l		old_int2,(0x00000068)
		move.l		old_int3,(0x0000006C)
		move.l		old_int4,(0x00000070)
		move.l		old_int5,(0x00000074)
		move.l		old_int6,(0x00000078)
		move.l		old_int7,(0x0000007C)
		move.l		old_trap13,(0x000000b4)
		move.l		old_trap14,(0x000000b8)
		bset.b		#2,(0x00600001)
		trap		#1
		move.b		#0xb2,(0x00600017)" : : : "d0");

      UnrelocateFile(BootFileContext->filehandle);
      HeapFree(BootFileContext->header.id);

      PortRestore();
    }
  }
  else {
  //
  // CODE TO TELL THE USER THAT proboot DIDN'T EXIST?
  //
  }
}
