/******************
* EVENT FUNCTIONS *
******************/

// This is where we'll end up when the event has finished. We now
// need to restore whatever state the process was in.
void ReturnFromEvent(void)
{
  CurrentProcessContext->flags.bits.in_interrupt = 0;

  // If the process is sleeping it hasn't been woken up by the event,
  // so we return it to the SleepingContext list.
  if(CurrentProcessContext->flags.bits.sleeping) {
    ToggleProsit();
    ListRemove(&ContextList, &CurrentProcessContext->header);
    ListInsert(&SleepingContextList, &CurrentProcessContext->header);
    ToggleProsit();
  }

  // Restore the process to the state it was before the event was triggered.
  asm("	unlk		%a6
	trap		#13
	subq.l		#6,%sp
	movem.l		%a1,-(%sp)
	move.l		CurrentProcessContext,%a1
	bclr.b		#1,8+16*4+6+9*4+1(%a1)
	move.l		%usp,%a1		/* USP used in event */
	addq.l		#2,%a1			/* Skip over triggering process */
	move.w		(%a1)+,4(%sp)		/* Restore SR */
	move.l		(%a1)+,4+2(%sp)		/* Restore PC */
	move.l		%a1,%usp		/* Restore USP */
	movem.l		(%sp)+,%a1
	rte");
}

//
// Trigger an event in another process. It will begin executing the
// event code when it gets the CPU the next time.
//
// NOTE: The redefinition MUST be here to get the right arguments - it
//  differs from the one presented to the programs!
BOOL TriggerEvent(HANDLE target, ULONG event)
{
  void *header;

  if((header = GetPrositHeaderOfProcess(target))) {
    while((header = FindPrositHeaderField(header, PROSIT_EVENT)) &&
          (((PrositHeaderElement_Event *)header)->event != event));

    if(header != NULL) {
      volatile PROCESS_CONTEXT_ENTRY *Context = (PROCESS_CONTEXT_ENTRY *)HeapDeref(target);
      void *FileStart = HeapDeref(Context->filehandle);
      void *FileEnd = FileStart+((PROSIT_FILE_HEADER *)FileStart)->filesize;

      // If another interrupt is running we'll wait for it to finish.
      while(Context->flags.bits.in_interrupt)
        BreakAndCooperate();

      Context->flags.bits.in_interrupt = 1;

      // Wait for process to return from an eventual library call.
      while((Context->pc < FileStart) || (Context->pc > FileEnd))
        BreakAndCooperate();

      // Turn off the system, redirect the PC of the target process,
      // (making sure it returns to an address of our choice when its
      // events is done) and turn the system back on.
      ToggleProsit();
      asm("		move.l		(%0),%%a2			/* Old USP */
			move.l		(%1),-(%%a2)			/* Old PC */
			move.w		%2,-(%%a2)			/* Old SR */
			move.w		" XSTR(CurrentProcessID) ",-(%%a2)	/* Current PID */
			move.l		#ReturnFromEvent,-(%%a2)	/* Return to here when done */
			move.l		%%a2,(%0)			/* New USP */
			move.l		%3,(%1)"			/* New PC */
        : : "a" (&Context->usp), "a" (&Context->pc), "d" (Context->sr),
            "a" (((PrositHeaderElement_Event *)header)->code) : "a2");
      ToggleProsit();

      if(Context->flags.bits.sleeping)
        WakeUp(target);

      return(TRUE);
    }
    else
      return(FALSE);
  }
  else
    return(FALSE);
}

//
// Trigger an event in all process except the calling one. They will
// begin executing the event code when they get the CPU the next time.
//
// NOTE: The redefinition MUST be here to get the right arguments - it
//  differs from the one presented to the programs!
void BroadcastEvent(ULONG event)
{
  ListElementHeader *current = CurrentProcessContext->header.next;
  int i;

  for(i=ListSize(&ContextList)-1 ; i>0 ; i--) {
//    TriggerEvent(current->id, event);
    current = current->next;
  }
}
