/*****************
* FILE FUNCTIONS *
*****************/

//
// Relocate a file and return the filehandle.
// If the file is archived the handle of the RAM copy will be returned (see inside).
// Returns NULL if: the file wasn't an executable
//                  the relocation failed
//                  there wasn't enough memory for a RAM copy
//                  the file was already relocated
//
PROSIT_FILE_HEADER *RelocateFile(HANDLE filehandle)
{
  PROSIT_FILE_HEADER *FileStart;

  // Check to see that:
  // * the filehandle is valid
  // * the filetype is 0xf3 (asm)
  // * the signature is "68kP"
  // * the file isn't already relocated
  if((FileStart = (PROSIT_FILE_HEADER *)HeapDeref(filehandle)) &&
     (((MEMORY_BYTE *)FileStart+FileStart->filesize+1)->data == 0xf3) &&
      (FileStart->signature == 0x36386b50) &&
      (FileStart->reloccount == 0)) {

    WORD reloc_result;

    // Check to see if the file is archived. If it is we must allocate memory and
    // copy it to RAM for execution. It is important to notice that the original
    // archived variable is not touched anymore after it has been copied.
    if((ULONG)FileStart > 0x200000)
      if((filehandle = HeapAlloc((ULONG)FileStart->filesize)))
      {
        FileStart = (PROSIT_FILE_HEADER *)memcpy((void *)HeapDeref(filehandle),(void *)FileStart,(ULONG)FileStart->filesize);
        FileStart->flags.bits.copy_of_archived = 1;
      }

    // Now to deal with the actual relocation.
    asm(" move.w %1,%%d0\n move.l (0x38),%%a0\n jsr (%%a0)" :
        "=d" (reloc_result) : "d" (filehandle) : "a0");
    if(reloc_result != 0)
      FileStart = NULL;
  }

  return(FileStart);
}

//
// Unrelocate a file.
// If the file was a copy of an archived file it will be deleted
// after it's been unrelocated.
//
void UnrelocateFile(HANDLE filehandle)
{
  asm(" move.w %0,%%d0\n move.l (0x40),%%a0\n jsr (%%a0)" : : "d" (filehandle) : "a0");
  if(((PROSIT_FILE_HEADER *)HeapDeref)->flags.bits.copy_of_archived)
    HeapFree(filehandle);
}

//
// Execute a non-prosit-executable file
//
void ExecuteFile(HANDLE filehandle)
{
  PROSIT_FILE_HEADER *FileStart;

  ToggleProsit();

  if(FileStart = RelocateFile(filehandle)) {
    asm("	movem.l		%%d0-%%d7/%%a0-%%a6,-(%%sp)
		move.l		%0,%%a0
		jsr		(%%a0)
		movem.l		(%%sp)+,%%d0-%%d7/%%a0-%%a6"
      : : "a" ((void *)FileStart+FileStart->offset_main+2));
    UnrelocateFile(filehandle);
  }

  ToggleProsit();
}
