/********************
* PROCESS FUNCTIONS *
********************/
PROCESS_CONTEXT_ENTRY *CreateProcess(void *header, void *entrypoint)
{
  PrositHeaderElement_Stacksize *stacksizefield;
  PROCESS_CONTEXT_ENTRY *Context;

  if((stacksizefield = FindPrositHeaderField(header,PROSIT_STACKSIZE)) &&
     (((PrositHeaderElement_Version *)(FindPrositHeaderField(header,PROSIT_VERSION)))->version == PROSIT_VERSION_STRING_N) &&
     (((PrositHeaderElement_Filetype *)(FindPrositHeaderField(header,PROSIT_FILETYPE)))->filetype == PROSIT_FILETYPE_PROGRAM_N) &&
     (HeapAvail() > (sizeof(PROCESS_CONTEXT_ENTRY) + stacksizefield->stacksize))) {

    WORD ProcessID;
    MEMORY_LONG *stack1;
    MEMORY_WORD *stack2;

    Context = HeapDeref(ProcessID = HeapAlloc(sizeof(PROCESS_CONTEXT_ENTRY)));

    Context->header.id = ProcessID;
    Context->header.priority = 0;
    ListInsert(&ContextList,&Context->header);

    Context->usp = HeapDeref(Context->stackhandle = HeapAlloc(stacksizefield->stacksize)) + stacksizefield->stacksize-10;
    stack1 = (void *)Context->usp;
    stack2 = (void *)Context->usp+8;
    stack1->data = (ULONG)TerminateProcess;
    stack2->data = Context->header.id;

    Context->sr = 0;
    Context->pc = entrypoint;
    Context->prositheader = header;

    asm(" movem.l	(0x00007464),%%a4-%%a5/%%d0/%%d2-%%d7
          movem.l	%%a4-%%a5/%%d0/%%d2-%%d7,8+16*4+6(%0)"
      : : "a" (Context) : "a4", "a5", "d0", "d2", "d3", "d4", "d5", "d6", "d7");
  }
  else
    Context = NULL;
  return(Context);
}

//
// Terminate a process... this is also where a self-terminated
// process will end up.
//
void TerminateProcess(HANDLE pid)
{
  PROCESS_CONTEXT_ENTRY *Context, *next;
  ListHeader *list = &ContextList;

  ToggleProsit();

  if(!(Context = (PROCESS_CONTEXT_ENTRY *)ListFind(&ContextList, pid)))
    Context = (PROCESS_CONTEXT_ENTRY *)ListFind((list = &SleepingContextList), pid);

  if(Context != NULL) {
    Context = HeapDeref(pid);
    next = Scheduler();
    ListRemove(list,&Context->header);
    HeapFree(Context->stackhandle);
    HeapFree(pid);

    // If the calling process is the same as the terminated process we must
    // jump into the scheduler.
    if(pid == CurrentProcessID)
      asm(" move.l %0,%%a0\n bra PostScheduler" : : "a" (next));
  }

  ToggleProsit();
}

//
// Create a process and add it to the ContextList.
// Returns the forks ID, NULL if not error.
//
HANDLE NewProcess(HANDLE filehandle)
{
  PROCESS_CONTEXT_ENTRY *Context;
  void *header;
  PROSIT_FILE_HEADER *FileStart;

  ToggleProsit();

  if((header = GetPrositHeaderOfFile(filehandle)) &&
     (FileStart = RelocateFile(filehandle)) &&
     (Context = CreateProcess(header, (void *)FileStart+FileStart->offset_main+2))) {
    Context->flags.all = CS_INIT_NORMAL;
    Context->filehandle = filehandle;
    ToggleProsit();
    return(Context->header.id);
  }
  else {
    ToggleProsit();
    return(NULL);
  }
}

//
// Create a fork and add it to the ContextList.
// Returns the forks ID, NULL if not error.
//
HANDLE ForkProcess(void *header, void *entrypoint)
{
  PROCESS_CONTEXT_ENTRY *Context;

  ToggleProsit();
  if((Context = CreateProcess(header, entrypoint))) {
    PROCESS_CONTEXT_ENTRY *ParentContext = (PROCESS_CONTEXT_ENTRY *)HeapDeref(CurrentProcessID);

    Context->flags.all = (ParentContext->flags.all & CS_FORK_KEEP_FLAGS) | CS_INIT_FORK;
    Context->filehandle = ParentContext->filehandle;
  }

  ToggleProsit();
  return(Context->header.id);
}

// Put a process to sleep, i.e. move it from ContextList
// to SleepingContextList.
void Sleep(HANDLE pid)
{
  PROCESS_CONTEXT_ENTRY *Context, *next;

  if((Context = (PROCESS_CONTEXT_ENTRY *)ListFind(&ContextList, pid))) {
    ToggleProsit();
    Context->flags.bits.sleeping = 1;
    next = Scheduler();
    ListRemove(&ContextList, &Context->header);
    ListInsert(&SleepingContextList, &Context->header);

    if(pid == CurrentProcessID) {
      asm("	move.w		%%sr,-(%%sp)
		or.w		#0x0700,%%sr
		move.l		%%a6,-(%%sp)
		move.l		CurrentProcessContext,%%a6
		movem.l		%%d0-%%d7/%%a0-%%a5,8(%%a6)
		move.l		(%%sp)+,%%a5
		move.l		(%%a5)+,8+14*4(%%a6)
		move.l		(%%a5)+,8+16*4+2(%%a6)
		addq.l		#2,%%a5
		move.l		%%a5,8+15*4(%%a6)
		move.w		(%%sp)+,8+16*4(%%a6)
		movem.l		(0x00007464),%%a0/%%d0-%%d7
		movem.l		%%a0/%%d0-%%d7,8+16*4+6(%%a6)
		move.l		%0,%%a0
		bra		PostScheduler" : : "a" (next));
    }
    ToggleProsit();
  }
}




// Wake-up a sleeping process.
void WakeUp(HANDLE pid)
{
  PROCESS_CONTEXT_ENTRY *Context;

  if((Context = (PROCESS_CONTEXT_ENTRY *)ListFind(&SleepingContextList, pid))) {
    ToggleProsit();

    ListRemove(&SleepingContextList, &Context->header);
    ListInsert(&ContextList, &Context->header);

    ToggleProsit();
  }
  else if(pid = CurrentProcessID) {
    Context->flags.bits.sleeping = 0;
  }
}
