#define TI92PLUS

#include <romcalls.h>
#include <prosit.h>
#include <winlib.h>
#include <wmlib.h>
#include <gmlib.h>
int _library;

#ifdef TI92PLUS
  WORD MouseDirectionData[16][2] = {{0,0},{0,0},{0,0},{1,1},{0,0},{0,0},{-1,1},{0,1},
                                    {0,0},{1,-1},{0,0},{1,0},{-1,-1},{0,-1},{-1,0},{0,0}};
#endif

#ifdef TI89
  WORD MouseDirectionData[16][2] = {{0,0},{0,0},{0,0},{-1,-1},{0,0},{0,0},{1,-1},{0,-1},
                                    {0,0},{-1,1},{0,0},{-1,0},{1,1},{0,1},{1,0},{0,0}};
#endif

#define LCD_ADDRESS 0x4c00
#define MOUSE_DELAYS 31

// Constants
WORD LeftMasks[16] = {0xffff, 0x7fff, 0x3fff, 0x1fff, 0x0fff, 0x07ff, 0x03ff, 0x01ff,
                      0x00ff, 0x007f, 0x003f, 0x001f, 0x000f, 0x0007, 0x0003, 0x0001};
WORD RightMasks[16] = {0x8000, 0xc000, 0xe000, 0xf000, 0xf800, 0xfc00, 0xfe00, 0xff00,
                       0xff80, 0xffc0, 0xffe0, 0xfff0, 0xfff8, 0xfffc, 0xfffe, 0xffff};

int MouseDelays[MOUSE_DELAYS] = {8, 6, 5, 4, 4, 4, 4, 4, 4, 4, 3, 3, 3,
                       3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 0};


// Exported variables
volatile int MouseX = 0;
volatile int MouseY = 0;
PWINDOW BackgroundWindow = EMPTYPWINDOWSTRUCT;
PWINDOW *WindowInFocus = &BackgroundWindow;
WIN_RECT BackgroundWindowRect = {0,0,LCD_WIDTH-1,LCD_HEIGHT-1};

// Initialized global variables




enum MouseStates {M_DISABLED=0, M_INACTIVE=1, M_ACTIVE=2, M_PLOTTING=3};

// Uninitialized global variables are allocated on the BSS!
ULONG OldInt5Address;

// Greyscale variables
HANDLE Plane2Handle;
void *Plane2;
WORD GreyscaleData[3];
WORD WLI_phase;

// Mouse variables
int MouseDelay, CurrentMouseDelayIndex;
volatile ULONG MousePlotMask;
int MouseHotspotX, MouseHotspotY;
volatile int MouseVisibleRows;
int MouseHeight, MouseWidth, MousePlotX, MousePlotY;
int MouseGFXOffset;
HANDLE MouseID, MousebuttonID;
volatile BYTE MouseState;
ULONG MouseLCDAndMaskBuffer[3][16];
// ULONG MouseShiftedMaskBuffer[16];
ULONG MouseWindowRefreshBuffer[16];
void *MouseGFX;
volatile int MouseLCDOffset;
volatile WORD mmx;
volatile WORD mmy;

ListHeader WindowList;




/*********************
*                    *
* INTERNAL FUNCTIONS *
*                    *
*********************/

void CreateAllMasks(void)
{
  void *BackgroundMask = HeapDeref(BackgroundWindow.MaskHandle) + BackgroundWindow.WidthInBytes + 4;
  PWINDOW *Current = (PWINDOW *)WindowList.start;

  // Fill in the background window mask.
#ifdef TI92PLUS
  asm("		moveq		#-1,%%d0
	CAMCBLoop:
		move.l		%%d0,(%1)+
		move.l		%%d0,(%1)+
		move.l		%%d0,(%1)+
		move.l		%%d0,(%1)+
		move.l		%%d0,(%1)+
		move.l		%%d0,(%1)+
		move.l		%%d0,(%1)+
		move.w		%%d0,(%1)+
		clr.l		(%1)+
		dbra		%0,CAMCBLoop" : :
               "d" (LCD_HEIGHT - 1), "a" (BackgroundMask) : "1", "d0");
#endif
#ifdef TI89
  asm("		moveq		#-1,%%d0
	CAMCBLoop:
		move.l		%%d0,(%1)+
		move.l		%%d0,(%1)+
		move.l		%%d0,(%1)+
		move.l		%%d0,(%1)+
		move.l		%%d0,(%1)+
		clr.l		(%1)+
		dbra		%0,CAMCBLoop" : :
               "d" (LCD_HEIGHT - 1), "a" (BackgroundMask) : "1", "d0");
#endif

  while(Current != &BackgroundWindow) {
    void *Mask = HeapDeref(Current->MaskHandle);

    asm("	CAMRLoop:
			move.l	%0,%%a5		/* Start of preliminary mask */
			move.w	%3,%%d6
		CAMCLoop:
			move.w	(%%a5)+,%%d7	/* Window mask */
			and.w	(%1),%%d7	/* Add LCD mask */
			move.w	%%d7,(%2)+	/* Store new window mask */
			eor.w	%%d7,(%1)+	/* Remove window mask from LCD mask */
			dbra	%%d6,CAMCLoop

			lea	0(%1,%5.w),%1	/* Next LCD mask row */
			dbra	%4,CAMRLoop" : :
		"a" (Mask),
		"a" ((void *)BackgroundMask + Current->LCDMaskStartOffset),
		"a" ((void *)Mask + (Current->WidthInBytes + 4)),
		"d" (((Current->WidthInBytes + 4) >> 1) - 1),
		"d" (Current->VisibleRows - 1),
                "d" ((LCD_WIDTH/8) - Current->WidthInBytes) : "1", "2", "4", "a5", "d6", "d7");

    Current = (PWINDOW *)Current->header.next;
  }
}



void CalculateWindow(PWINDOW *w)
{
  int i;
  WORD StartX, EndX, StartXOffset;
  void *Mask = HeapDeref(w->MaskHandle);

  /**********************
  * THE SHIFT CONSTANTS *
  **********************/

  //
  // Deal with Y
  //
  if(w->Y < 0) {
    // Case 1: Window cut off by upper edge of screen.
    w->VisibleRows = w->Height + w->Y;
    w->OffsetInContents = w->WidthInBytes * (-w->Y);
    w->LCDStartOffset = 0;
    w->LCDMaskStartOffset = 0;
  }
  else if((w->Y + w->Height) > LCD_HEIGHT) {
    // Case 2: Window cut off by lower edge of screen.
    w->VisibleRows = LCD_HEIGHT - w->Y;
    w->OffsetInContents = 0;
    w->LCDStartOffset = 30*w->Y;
    w->LCDMaskStartOffset = ((LCD_WIDTH/8) + 4)*w->Y;
  }
  else {
    // Case 3: Entire window height fits on-screen.
    w->VisibleRows = w->Height;
    w->OffsetInContents = 0;
    w->LCDStartOffset = 30*w->Y;
    w->LCDMaskStartOffset = ((LCD_WIDTH/8) + 4)*w->Y;
  }

  //
  // Deal with X
  //
  if(w->X < 0) {
    w->LCDStartOffset -= 2*((15 - w->X) >> 4);
    w->LCDMaskStartOffset -= 2*((15 - w->X) >> 4);
  }
  else {
    w->LCDStartOffset += 2*(w->X >> 4);
    w->LCDMaskStartOffset += 2*(w->X >> 4);
  }


  /***********
  * THE MASK *
  ***********/

  // Get start X of mask.
  if(w->X < 0)
    StartX = (15-w->X) & 0xfff0;
  else
    StartX = w->X & 0x000f;

  // Get end X of mask.
  if((w->X + w->Width) > LCD_WIDTH)
    EndX = StartX + LCD_WIDTH - w->X - 1;
  else
    EndX = (w->X & 0x000f) + w->Width - 1;

  // Get offset and no. of columns.
  StartXOffset = StartX >> 4;
  w->VisibleColumns = (EndX >> 4) - StartXOffset + 1;

  // Clear the mask.
  for(i=0 ; i <= w->WidthInBytes ; i+=2)
    asm(" clr.w 0(%0,%1.w)" : : "a" (Mask), "d" (i));

  // Set the left mask.
  asm("	lea	0(%0,%1.w),%0
	move.w	%2,(%0)+" : :
       "a" (Mask), "d" (StartXOffset << 1), "d" (LeftMasks[StartX & 0x000f]));

  // Fill in the middle mask.
  for(i=2 ; i < w->VisibleColumns ; i++) {
    asm("move.w #0xffff,(%0)+" : : "a" (Mask));
  }

  // Set the right mask.
  asm("	move.w	%1,(%0)" : : "a" (Mask), "d" (RightMasks[EndX & 0x000f]));
}


void RefreshScreen(void)
{
  ListElementHeader *Current;
  int i;

  CreateAllMasks();

  Current = WindowList.start;
  for(i=0 ; i<ListSize(&WindowList) ; i++) {
    PWinRefresh((PWINDOW *)Current);
    Current = Current->next;
  }
}

/******************
* Mouse functions *
******************/
#include "winlib_mouse.h"

/************************
* Mousebutton functions *
************************/
#include "winlib_mousebutton.h"

//
// The winlib auto-interrupt 5 hook. Calls MouseInterrup_Function
// and manages greyscale.
//
asm("	WinlibInterrupt:
		movem.l		%a0-%a2/%d0-%d2,-(%sp)
		bsr		MouseInterrupt_Function

		move.w		#0x2700,%sr
		move.w		WLI_phase,%d0
		lea		GreyscaleData,%a0
		move.w		0(%a0,%d0.w),(0x00600010)
		subq.w		#2,%d0
		bge		WLI_no_wrap
		moveq		#2*2,%d0
	WLI_no_wrap:
		move.w		%d0,WLI_phase
	WLI_skip:

		movem.l		(%sp)+,%a0-%a2/%d0-%d2
		move.l		OldInt5Address,-(%sp)
		rts");



BOOL InitWindowSystem(MOUSEPOINTER *mp)
{
  int i;

  // Setup the greyscale. If there's not enough memory we return FALSE.
  if(Plane2Handle = HeapAlloc(30*128 + 8 + 10)) {
    Plane2 = (void *)(((ULONG)HeapDeref(Plane2Handle) + 7) & 0xfffffff8);
    GreyscaleData[0] = ((ULONG)Plane2 >> 3);
    GreyscaleData[1] = GreyscaleData[2] = (LCD_ADDRESS >> 3);
    WLI_phase = 0;

    // Set the WindowInFocus.
    WindowInFocus = &BackgroundWindow;
    ListInitialize(&WindowList);

    // Initialize the mouse variables
    MouseDelay = 0;
    CurrentMouseDelayIndex = 0;
    MouseWidth = mp->Width;
    MouseHeight = mp->Height;
    MouseX = LCD_WIDTH/2;
    MouseY = LCD_HEIGHT/2;
    MouseHotspotX = mp->HotSpotX;
    MouseHotspotY = mp->HotSpotY;
    MouseGFX = &mp->GFX[0];

    MouseState = M_DISABLED;
    ActivateMouse();

    // Hook the mouse coordinate interrupt and the greyscale interrupt.
    asm("	move.w		#0x0700,%%d0
		trap		#1
		move.l		(0x00000074),OldInt5Address
		bclr.b		#2,(0x00600001)
		move.l		#WinlibInterrupt,(0x00000074)
		bset.b		#2,(0x00600001)
		trap		#1" : : : "d0");

    asm("	move.b		#0xf1,(0x00600017)");

    // Start Mouse and Mousebutton processes
    MouseID = ForkProcess(&Mouse_Header, &Mouse_Process);
    MousebuttonID = ForkProcess(&Mousebutton_Header, &Mousebutton_Process);

    // Wait for Mouse process to start
    while(MouseState != M_ACTIVE) BreakAndCooperate();

    // Open background window
    BackgroundWindow.Win.Flags = WF_NOBORDER;
    BackgroundWindow.header.flags = 0;
    BackgroundWindow.header.priority = -127;
    PWinOpen(&BackgroundWindow, &BackgroundWindowRect, NULL, NULL);

    return(TRUE);
  }
  else
    return(FALSE);
}


void CloseWindowSystem(void)
{
  ListRemove(&WindowList, &BackgroundWindow.header);
  HeapFree(BackgroundWindow.ContentsHandle[0]);
  HeapFree(BackgroundWindow.ContentsHandle[1]);
  HeapFree(BackgroundWindow.MaskHandle);

  TerminateProcess(MouseID);
  TerminateProcess(MousebuttonID);

  // Unhook the mouse coordinate interrupt and the greyscale interrupt.
  asm("	move.w		#0x0700,%%d0
	trap		#1
	bclr.b		#2,(0x00600001)
	move.l		OldInt5Address,(0x00000074)
	bset.b		#2,(0x00600001)
	trap		#1" : : : "d0");

  // Restore the LCD pointer to LCD_ADDRESS
  asm("	move.w		GreyscaleData+2,(0x00600010)");

  HeapFree(Plane2Handle);
}

BOOL PWinOpen(PWINDOW *w, WIN_RECT *dimensions, PWINDOW_DATARETURN *wr, char *title)
{
  ULONG WindowSize, MaskSize;

  // Set the window position and size
  w->X = dimensions->x0;
  w->Y = dimensions->y0;

  w->Width = dimensions->x1 - dimensions->x0 + 1;
  w->Height = dimensions->y1 - dimensions->y0 + 1;

  w->WidthInBytes = ((((w->Width - 1) >> 4) + 1) << 1);

  WindowSize = (w->WidthInBytes * w->Height) + 4;
  MaskSize = ((w->WidthInBytes + 4) * (w->Height + 1));

  if(HeapAvail() > (2*WindowSize + MaskSize)) {
    BITMAP *Contents1, *Contents2;

    // Allocate memory for window contents and mask.
    Contents1 = (BITMAP *)HeapDeref(w->ContentsHandle[0] = HeapAllocHigh(WindowSize));
    Contents2 = (BITMAP *)HeapDeref(w->ContentsHandle[1] = HeapAllocHigh(WindowSize));
    w->MaskHandle = HeapAllocHigh(MaskSize);

    // Set the content BITMAP dimensions.
    Contents1->NumRows = Contents2->NumRows = w->Height;
    Contents1->NumCols = Contents2->NumCols = w->WidthInBytes;

    // Set some TIOS WINDOW data.
    w->Win.Flags &= WF_TTY + WF_ACTIVE + WF_NOBORDER + WF_NOBOLD + WF_TITLE + WF_DIRTY +
                    WF_VISIBLE;
    w->Win.Flags |= WF_DONT_REALLOC + WF_VIRTUAL + WF_DUP_ON + WF_DUP_SCR;
    w->Win.CurFont = 0;
    w->Win.CurAttr = A_NORMAL;
    w->Win.Background = A_REVERSE;
    w->Win.TaskId = CurrentProcessID;
    w->Win.Client.xy.x0 = w->Win.Client.xy.y0 = 0;
    w->Win.Client.xy.x1 = w->Width - 1;
    w->Win.Client.xy.y1 = w->Height - 1;
    w->Win.Port = w->Win.Clip = w->Win.Client;
    w->Win.Port.xy.x1 = (w->WidthInBytes << 3) - 1;
    if(w->Win.Title = title)
      w->Win.Flags |= WF_TITLE;

    // Clear the window.
    PWinSelectPlane(w, 1);
    WinClr(&w->Win);
    PWinSelectPlane(w, 0);
    WinClr(&w->Win);

    // Initialize the window datareturn area if it exists.
    if((w->Datareturn = wr)) {
      wr->Key = wr->ButtonID = 0;
      wr->w = NULL;
    }

    // Prepare the window and insert it into the window list.
    CalculateWindow(w);
    ListInsert(&WindowList, &w->header);
    CreateAllMasks();

    // Remove focus from whatever window had it before.
    WM_WindowLostFocus(WindowInFocus);

    // Initialize and give focus to the new window.
    WM_InitializeWindow(w);
    GM_InitializeWindow(w);
    WM_WindowGotFocus(w);
    WindowInFocus = w;

/*
    // Draw the GUI structure if it exists.
    if((w->StructPtr = wg) && (wg->start)) {
      GM_DrawStructure(w, wg);
    }
*/

    PWinRefresh(w);

    return(TRUE);
  }
  else
    return(FALSE);
}

void PWinClose(PWINDOW *w)
{
  ListRemove(&WindowList, &w->header);
  HeapFree(w->ContentsHandle[0]);
  HeapFree(w->ContentsHandle[1]);
  HeapFree(w->MaskHandle);
  RefreshScreen();
}

void PWinRefresh(PWINDOW *w)
{
  int MouseMaskPlotX;
  volatile int StartRowInMouse, AffectedRows, MouseStartRow, MouseStartColumn;
  volatile void *Mask, *Contents1, *Contents2;

  // The mouse can't be allowed to move while we're refreshing the window.
  DeactivateMouse();

  MouseMaskPlotX = (MousePlotX & 0xfff0);

  // Add the mousepointer mask to the window (if mouse is over window).
  if(AffectedRows = ((MouseMaskPlotX < (w->X + w->Width)) &&
                     (MouseMaskPlotX > (w->X - 32)) &&
                     (MousePlotY < (w->Y + w->Height)) &&
                     (MousePlotY > (w->Y - MouseHeight)))) {

    MouseStartRow = MousePlotY - w->Y;
    AffectedRows = MouseHeight;
    StartRowInMouse = 0;

    if(MouseStartRow < 0) {
      AffectedRows += MouseStartRow;
      StartRowInMouse = -MouseStartRow;
      MouseStartRow = 0;
    }
    if((MouseStartRow + MouseHeight) > w->Height) {
      AffectedRows -= (MouseStartRow + MouseHeight - w->Height);
    }

//    MouseStartInMask = ((w->WidthInBytes + 2) * (MouseStartInMask + 1));

    if(MouseMaskPlotX < 0)
      MouseStartColumn = -(((15-MouseMaskPlotX) >> 4) << 1);
    else
      MouseStartColumn = ((MouseMaskPlotX >> 4) << 1);

    if(w->X < 0)
      MouseStartColumn += (((15-w->X) >> 4) << 1);
    else
      MouseStartColumn -= ((w->X >> 4) << 1);

/*
    if((MouseMaskPlotX - w->X) < 0)
      MouseStartColumnInContents = -(((15-(MouseMaskPlotX - w->X)) >> 4) << 1);
    else
      MouseStartColumnInContents = (((MouseMaskPlotX - w->X) >> 4) << 1);

    MouseStartColumnInContents = MouseStartColumnInMask - 2;
*/

    Mask = (void *)HeapDeref(w->MaskHandle) +
            ((w->WidthInBytes + 4) * (MouseStartRow + 1)) + MouseStartColumn;
    Contents1 = (void *)HeapDeref(w->ContentsHandle[0]) + 4 +
             (w->WidthInBytes * MouseStartRow) + MouseStartColumn - 2;
    Contents2 = (void *)HeapDeref(w->ContentsHandle[1]) + 4 +
             (w->WidthInBytes * MouseStartRow) + MouseStartColumn - 2;
;


    asm("	RWAMMLoop:
			move.l	(%2),%%d5	/* Old window mask */
			move.l	%%d5,(%1)+	/* Store old window mask */

			moveq	#0,%%d6		/* Get plane1 window contents */
			move.w	(%3),%%d6
			ror.l	%5,%%d6
			clr.w	%%d6
			move.l	2(%3),%%d7
			lsr.l	%5,%%d7
			or.l	%%d6,%%d7

			and.l	%%d5,%%d7	/* Mask to get affected LCD buffer */
			not.l	%%d5
			and.l	%%d5,(%0)	/* Mask to get unaffected LCD buffer */
			or.l	%%d7,(%0)+	/* Add new LCD buffer part */

			moveq	#0,%%d6		/* Get plane2 window contents */
			move.w	(%4),%%d6
			ror.l	%5,%%d6
			clr.w	%%d6
			move.l	2(%4),%%d7
			lsr.l	%5,%%d7
			or.l	%%d6,%%d7

			and.l	%%d5,(%0)	/* Mask to get unaffected LCD buffer */
			not.l	%%d5
			and.l	%%d5,%%d7	/* Mask to get affected LCD buffer */
			or.l	%%d7,(%0)+	/* Add new LCD buffer part */

			move.l	(%0)+,%%d6	/* Shifted mouse mask */
			and.l	(%2),%%d6	/* Old window mask AND mouse mask */
			move.l	%%d6,(%2)	/* Store new mouse mask */

			lea	4(%2,%7.w),%2	/* Next row in window mask */
			lea	(%3,%7.w),%3	/* Next row in plane1 contents */
			lea	(%4,%7.w),%4	/* Next row in plane2 contents */
			dbra	%6,RWAMMLoop" : :
		"a" ((void *)&MouseLCDAndMaskBuffer + 12*StartRowInMouse),
		"a" (&MouseWindowRefreshBuffer),
		"a" (Mask),
		"a" (Contents1),
		"a" (Contents2),
		"d" (w->X & 0x000f),
		"d" (AffectedRows - 1),
		"d" (w->WidthInBytes) : "1", "2", "6", "d5", "d6", "d7");
  }

  // Plot the first plane.
  asm("		movem.l		%0/%1/%2/%3/%%d5/%%d6/%%d7,-(%%sp)
	RW1RowLoop:
		moveq		#0,%%d6
		move.w		%4,%%d5			/* No. of columns in window */
	RW1ColLoop:
		moveq		#0,%%d7
		move.w		(%0)+,%%d7		/* GFX data */
		ror.l		%5,%%d7			/* Rotate GFX data */
		or.w		%%d6,%%d7		/* Add old GFX data */
		move.w		(%1)+,%%d6		/* Mask data */
		and.w		%%d6,%%d7		/* Mask GFX data */
		not.w		%%d6			/* Invert mask */
		and.w		(%2),%%d6		/* Get and mask old LCD GFX */
		or.w		%%d6,%%d7		/* Add LCD GFX to GFX */
		move.w		%%d7,(%2)+		/* Store new GFX data */
		swap		%%d7			/* Store the old upper word... */
		move.w		%%d7,%%d6		/* ...for the next iteration */
		dbra		%%d5,RW1ColLoop

		move.w		(%1)+,%%d6		/* Mask data */
		and.w		%%d6,%%d7		/* Mask GFX data */
		not.w		%%d6			/* Invert mask */
		and.w		(%2),%%d6		/* Get and mask old LCD GFX */
		or.w		%%d6,%%d7		/* Add LCD GFX to GFX */
		move.w		%%d7,(%2)+		/* Store new GFX data */

		addq.l		#2,%1			/* Skip rightmost mask word */

		lea		(%2,%6.w),%2
		dbra		%3,RW1RowLoop
		movem.l		(%%sp)+,%0/%1/%2/%3/%%d5/%%d6/%%d7" : :
              "a" ((void *)HeapDeref(w->ContentsHandle[0]) + w->OffsetInContents + 4),
              "a" ((void *)HeapDeref(w->MaskHandle) + (w->WidthInBytes + 4)),
              "a" ((void *)LCD_ADDRESS + w->LCDStartOffset),
              "d" (w->VisibleRows - 1),
              "d" ((w->WidthInBytes >> 1) - 1),
              "d" (w->X & 0x000f),
              "d" (28 - w->WidthInBytes));

  // Plot the second plane.
  asm("		movem.l		%0/%1/%2/%3/%%d5/%%d6/%%d7,-(%%sp)
	RW2RowLoop:
		moveq		#0,%%d6
		move.w		%4,%%d5			/* No. of columns in window */
	RW2ColLoop:
		moveq		#0,%%d7
		move.w		(%0)+,%%d7		/* GFX data */
		ror.l		%5,%%d7			/* Rotate GFX data */
		or.w		%%d6,%%d7		/* Add old GFX data */
		move.w		(%1)+,%%d6		/* Mask data */
		and.w		%%d6,%%d7		/* Mask GFX data */
		not.w		%%d6			/* Invert mask */
		and.w		(%2),%%d6		/* Get and mask old LCD GFX */
		or.w		%%d6,%%d7		/* Add LCD GFX to GFX */
		move.w		%%d7,(%2)+		/* Store new GFX data */
		swap		%%d7			/* Store the old upper word... */
		move.w		%%d7,%%d6		/* ...for the next iteration */
		dbra		%%d5,RW2ColLoop

		move.w		(%1)+,%%d6		/* Mask data */
		and.w		%%d6,%%d7		/* Mask GFX data */
		not.w		%%d6			/* Invert mask */
		and.w		(%2),%%d6		/* Get and mask old LCD GFX */
		or.w		%%d6,%%d7		/* Add LCD GFX to GFX */
		move.w		%%d7,(%2)+		/* Store new GFX data */

		addq.l		#2,%1			/* Skip rightmost mask word */

		lea		(%2,%6.w),%2
		dbra		%3,RW2RowLoop
		movem.l		(%%sp)+,%0/%1/%2/%3/%%d5/%%d6/%%d7" : :
              "a" ((void *)HeapDeref(w->ContentsHandle[1]) + w->OffsetInContents + 4),
              "a" ((void *)HeapDeref(w->MaskHandle) + (w->WidthInBytes + 4)),
              "a" ((void *)Plane2 + w->LCDStartOffset),
              "d" (w->VisibleRows - 1),
              "d" ((w->WidthInBytes >> 1) - 1),
              "d" (w->X & 0x000f),
              "d" (28 - w->WidthInBytes));

  // Remove the mousepointer mask from the window (if mouse was over window).
  if(AffectedRows) {
    asm("	RWRMMLoop:
			move.l	(%0)+,(%1)	/* Restore old window mask */
			lea	4(%1,%3.w),%1	/* Next row in window mask */
			dbra	%2,RWRMMLoop" : :
		"a" (&MouseWindowRefreshBuffer),
		"a" (Mask),
		"d" (AffectedRows - 1),
		"d" (w->WidthInBytes) : "0", "1", "2");
  }

  ActivateMouse();

  // Force the calling process to break to smooth things out when processes are
  // continuously calling RefreshWindow...
  BreakAndCooperate();
}

void PWinSelectPlane(PWINDOW *w, int plane)
{
  w->Win.DupScr = w->ContentsHandle[plane];
}

void PWinMoveTo(PWINDOW *w, int newx, int newy)
{
  w->X = newx;
  w->Y = newy;
  CalculateWindow(w);
  RefreshScreen();
}

void PWinMoveToFront(PWINDOW *w)
{
  ListRemove(&WindowList, &w->header);
  ListInsert(&WindowList, &w->header);

  CreateAllMasks();
  PWinRefresh(w);
}
