;---------------------------------------------------------------------
; Clear(WORD ID)
;
; Function: Clear the window contents block.
;---------------------------------------------------------------------
Clear:
winlib@000C:
	movem.l		a0/d0-d1,-(sp)

	move.w		12+4(sp),d0			;Get ID (handle) of window
	doorsos::DEREF	d0,a0
	move.w		W_WIDTH_IN_WORDS(a0),d0		;Width in words
	mulu.w		W_HEIGHT(a0),d0			;Width*Height in words
	subq.w		#1,d0
	move.w		W_CONTENTS_HANDLE(a0),d1	;Contents handle
	doorsos::DEREF	d1,a0				;*Contents
	addq.l		#4,a0

\cloop	clr.w		(a0)+				;Clear contents memory block
	dbra.s		d0,\cloop

	movem.l		(sp)+,a0/d0-d1
	rts

;---------------------------------------------------------------------
; SetClip(WORD ID, BYTE *clip)
;
; Function: Set window <ID> to use the clip area <clip>.
;---------------------------------------------------------------------
SetClip:
winlib@000D:
	movem.l		a0-a1/d0-d1,-(sp)

	move.w		4+4*4+0(sp),d0
	doorsos::DEREF	d0,a0		;*window
	move.l		4+4*4+2(sp),a1	;*clip

	move.w		(a1),d0		;x1
	move.w		4(a1),d1	;x2
	cmp.w		d1,d0
	ble.s		\xok
	exg.w		d1,d0		;x1<x2
\xok	tst.w		d0
	bge.s		\x1pos
	moveq		#0,d0		;x1 must be positive
\x1pos	cmp.w		W_WIDTH(a0),d1
	blt.s		\x2ok
	move.w		W_WIDTH(a0),d1
	subq.w		#1,d1		;x2 must be within the window!
\x2ok	move.b		d0,W_LEFT_CLIP(a0)
	move.b		d1,W_RIGHT_CLIP(a0)

	move.w		2(a1),d0	;y1
	move.w		6(a1),d1	;y2
	cmp.w		d1,d0
	ble.s		\yok
	exg.w		d1,d0		;y1<y2
\yok	tst.w		d0
	bge.s		\y1pos
	moveq		#0,d0		;y1 must be positive
\y1pos	cmp.w		W_HEIGHT(a0),d1
	blt.s		\y2ok
	move.w		W_HEIGHT(a0),d1
	subq.w		#1,d1		;y2 must be within the window!
\y2ok	move.b		d0,W_TOP_CLIP(a0)
	move.b		d1,W_BOTTOM_CLIP(a0)

\done	movem.l		(sp)+,a0-a1/d0-d1
	rts


;---------------------------------------------------------------------
; VertLine(WORD ID, WORD x, WORD y1, WORD y2, WORD flags,
;  [WORD pattern])
;
; Function: Draw a line from (x,y1) to (x,y2).
;
; Flags: 0 = Nothing
;        1 = Pattern
;        2 = White
;        3 = Black
;---------------------------------------------------------------------
VertLine:
winlib@000E:
	link		a6,#-8

	movem.l		a0/d0,-(sp)

	move.w		10(a6),-8(a6)	;x1
	move.w		12(a6),-6(a6)	;y1
	move.w		10(a6),-4(a6)	;x2
	move.w		14(a6),-2(a6)	;y2

	pea		-8(a6)		;*line
	move.w		8(a6),d0
	doorsos::DEREF	d0,a0
	move.l		a0,-(sp)
	jsr		doorsos::WinLine
	lea		8(sp),sp

	movem.l		(sp)+,a0/d0
	unlk		a6
	rts

;---------------------------------------------------------------------
; HorizLine(WORD ID, WORD y, WORD x1, WORD x2, WORD flags,
;  [WORD pattern])
;
; Function: Draw a line from (x1,y) to (x2,y).
;
; Flags: 0 = Nothing
;        1 = Pattern
;        2 = White
;        3 = Black
;---------------------------------------------------------------------
HorizLine:
winlib@000F:
	link		a6,#-8

	movem.l		a0/d0,-(sp)

	move.w		12(a6),-8(a6)	;x1
	move.w		10(a6),-6(a6)	;y1
	move.w		14(a6),-4(a6)	;x2
	move.w		10(a6),-2(a6)	;y2

	pea		-8(a6)		;*line
	move.w		8(a6),d0
	doorsos::DEREF	d0,a0
	move.l		a0,-(sp)
	jsr		doorsos::WinLine
	lea		8(sp),sp

	movem.l		(sp)+,a0/d0
	unlk		a6
	rts

;---------------------------------------------------------------------
; Box(WORD ID, WORD x, WORD y, WORD width, WORD height,
;  WORD flags, [WORD pattern], [WORD pattern], [WORD pattern],
;  [WORD pattern], [BYTE *gfx, BYTE *mask]
;
; Function: Draw a <width,height>-sized box at (x, y).
;---------------------------------------------------------------------
Box:
winlib@0010:
	movem.l		a0-a6/d0-d7,-(sp)

	move.w		4+15*4+0(sp),d0		;Window Handle

	move.w		d0,d7			;Store it
	doorsos::DEREF	d0,a0			;*Window
	move.w		W_CONTENTS_HANDLE(a0),d0
	doorsos::DEREF	d0,a2			;*Contents
	addq.l		#4,a2

	move.w		4+15*4+2(sp),d0		;x1
;	cmp.w		W_RIGHT_CLIP(a0),d0
;	bgt		\done
	move.w		4+15*4+4(sp),d1		;y1
;	cmp.w		W_BOTTOM_CLIP(a0),d1
;	bgt		\done
	move.w		4+15*4+6(sp),d2		;width
	add.w		d0,d2
	subq.w		#1,d2			;x2
;	cmp.w		W_LEFT_CLIP(a0),d2
;	blt		\done
	move.w		4+15*4+8(sp),d3		;height
	add.w		d1,d3
	subq.w		#1,d3			;y2
;	cmp.w		W_TOP_CLIP(a0),d3
;	blt		\done

	move.w		4+15*4+10(sp),d4	;Flags

	lea		4+15*4+12(sp),a1	;*Patterns on the stack

	lea		-12(sp),sp
	move.w		d7,(sp)
;Top line
	move.w		d2,6(sp)	;x2
	move.w		d0,4(sp)	;x1
	btst.w		#1,d4		;Line or pattern?
	bne.s		\tline		;Line...
	btst.w		#0,d4		;Pattern... use it?
	beq.s		\bot
	move.w		(a1)+,10(sp)	;Pattern
\tline	move.w		d4,8(sp)	;Flags
	move.w		d1,2(sp)	;y1
	bsr		HorizLine
	addq.w		#1,d1
;Bottom line
\bot	lsr.w		#2,d4
	btst.w		#1,d4		;Line or pattern?
	bne.s		\bline
	btst.w		#0,d4		;Pattern... use it?
	beq.s		\left
	move.w		(a1)+,10(sp)	;Pattern
\bline	move.w		d4,8(sp)	;Flags
	move.w		d3,2(sp)	;y2
	bsr		HorizLine
	subq.w		#1,d3
;Left line
\left	move.w		d3,6(sp)	;y2
	move.w		d1,4(sp)	;y1
	lsr.w		#2,d4
	btst.w		#1,d4		;Line or pattern?
	bne.s		\lline
	btst.w		#0,d4		;Pattern... use it?
	beq.s		\right
	move.w		(a1)+,10(sp)	;Pattern
\lline	move.w		d4,8(sp)	;Flags
	move.w		d0,2(sp)	;x1
	bsr		VertLine
	addq.w		#1,d0
;Right line
\right	lsr.w		#2,d4
	btst.w		#1,d4		;Line or pattern?
	bne.s		\rline
	btst.w		#0,d4		;Pattern... use it?
	beq.s		\donlin
	move.w		(a1)+,10(sp)	;Pattern
\rline	move.w		d4,8(sp)	;Flags
	move.w		d2,2(sp)	;x2
	bsr		VertLine
	subq.w		#1,d2

\donlin	lea		12(sp),sp

;*******************
;* INTERNAL OF BOX *
;*******************
	lsr.w		#2,d4
	cmp.b		#0,d4
	beq		\done		;No fill

;**************
;* RIGHT MASK *
;**************
\xdone	lea		wmasks(pc),a1
\done	movem.l		(sp)+,a0-a6/d0-d7
	rts


;---------------------------------------------------------------------
; PlotGFX(WINDOW *w, WORD x, WORD y, BYTE *gfx, BYTE *mask)
;
; Function: Plot <gfx> with mask <mask> at (x,y) in window <ID>, using
;  the window's current clip area. The format for gfx is described in
;  the docs.
;---------------------------------------------------------------------
PlotGFX:
winlib@0011:
	movem.l		a0-a6/d0-d7,-(sp)

	move.l		4+15*4+0(sp),a3		;*Window
	move.w		4+15*4+4(sp),d0		;x
	move.w		4+15*4+6(sp),d1		;y
	move.l		4+15*4+8(sp),a0		;*gfx
	move.l		4+15*4+12(sp),a6	;*mask

	moveq		#0,d2
	move.b		(a0)+,d2		;gfx_width (pixels)
	subq.w		#1,d2
	lsr.w		#3,d2
	addq.w		#1,d2			;gfx_width (bytes)
	moveq		#0,d3
	move.b		(a0)+,d3		;gfx_height (pixels)
	move.w		d2,a4			;Store gfx_width for later use

;	cmp.w		W_RIGHT_CLIP(a3),d0
	bgt		\exit			;Plot is to the right of clip, exit
	move.w		d2,d6
	lsl.w		#3,d6
	add.w		d0,d6
;	cmp.w		W_LEFT_CLIP(a3),d6
	ble		\exit			;Plot is to the left of clip, exit
;	cmp.w		W_BOTTOM_CLIP(a3),d1
	bgt		\exit			;Plot is below clip, exit
	move.w		d1,d7
	add.w		d3,d7
;	cmp.w		W_TOP_CLIP(a3),d7
	ble		\exit			;Plot is above clip, exit

	move.w		W_CONTENTS_HANDLE(a3),d6
	doorsos::DEREF	d6,a2
	addq.l		#4,a2
;	move.w		W_TOP_CLIP(a3),d6
	mulu.w		W_WIDTH_IN_WORDS(a3),d6
	lsl.w		#1,d6			;Since window width is in words
;	move.w		W_LEFT_CLIP(a3),d7
	lsr.w		#3,d7
	add.w		d7,d6
	add.w		d6,a2			;*plot

;****************
;* Y PARAMETERS *
;****************
;Lower Y edge
	move.w		d1,d6
	add.w		d3,d6			;y + gfx_height
	subq.w		#1,d6
;	sub.w		W_BOTTOM_CLIP(a3),d6		;y + gfx_height - y2_clip
	blt.s		\cont1
;Lower edge of plot exceeds lower clip limit, decrease no. of rows to plot
	sub.w		d6,d3

;Upper Y edge
\cont1	move.w		d1,d6
;	sub.w		W_TOP_CLIP(a3),d6	;y - y1_clip
	bge.s		\ypos
;GFX y starts before the clip, adjust GFX start parameters
	add.w		d6,d3			;Decrease no. of rows to plot
	muls.w		d2,d6			;(y - y1_clip)*gfx_width
	sub.w		d6,a0			;Move *gfx
	sub.w		d6,a6			;Move *mask
	bra.s		\cont2
;GFX y starts on or after the clip, adjust destination start parameters
\ypos	;mulu.w		W_WIDTH_IN_WORDS(a3),d6	;(y - y1_clip)*target_width
	lsl.w		#1,d6			;Since window wisth is in words
	add.w		d6,a2			;Move *plot

;
; Now d3 = No. of rows to plot
;     a0 = Where to start getting data, y-wise
;     a2 = Where to start plotting, y-wise
;

\cont2

;************************************
;* CREATE LEFT- AND RIGHTCLIP MASKS *
;************************************
;Create left bitmask
;	move.w		W_LEFT_CLIP(a3),d7
	cmp.w		d0,d7
	bge.s		\oklc
	move.w		d0,d7
\oklc	and.w		#%0000000000000111,d7
	lea		masks(pc),a5
	move.b		0(a5,d7.w),d6		;d6.uw = left mask
	swap		d6

;Create right bitmask
	move.w		d0,d7
	move.w		a4,d6
	lsl.w		#3,d6
	subq.w		#1,d7
	add.w		d6,d7			;x+width
;	cmp.w		W_RIGHT_CLIP(a3),d7
	ble.s		\okrc
	sub.w		#8,a5
;	move.w		W_RIGHT_CLIP(a3),d7
\okrc	and.w		#%0000000000000111,d7
	move.b		16(a5,d7.w),d6		;d6.lw = right mask

;****************
;* X PARAMETERS *
;****************
;Right X edge
	move.w		d0,d7
	asr.w		#3,d7			;x/8

;	move.w		W_RIGHT_CLIP(a3),d4
	lsr.w		#3,d4			;x2_clip/8
	sub.w		d7,d4			;(x2_clip/8) - (x/8)
	beq.s		\sp1

	move.w		#$4E71,\sm2		;NOP
	move.w		#$4E71,\sm21		;NOP
	cmp.w		d4,d2
	ble.s		\cont3
	bne.s		\ne
	subq.w		#1,d4

;(x/8)+width is on or to the right of (x2_clip/8)
\ne	move.w		#$1210,\sm2		;move.b (a0),d1
	move.w		#$1A16,\sm21		;move.b (a6),d5
	move.w		d4,d2
	bra.s		\cont3

;(x/8) is on the right clip byte, alter the masks and set the width to handle it
\sp1	moveq		#1,d2		;Set width to one to "fool" plot routine
	and.b		d6,d5		;Combine left and right masks into the left mask...
	move.b		#0,d6		;...and clear the old right mask

;Left X edge
\cont3	;move.w		W_LEFT_CLIP(a3),d4
	lsr.w		#3,d4			;x1_clip/8
	sub.w		d7,d4			;(x1_clip/8) - (x/8)
	ble.s		\xpos2
;(x/8) is to the left of (x1_clip/8), adjust GFX start parameters
	add.w		d4,a0			;Move *gfx
	add.w		d4,a6			;Move *mask
	sub.w		d4,d2			;Decrease no. of columns to plot
	bne.s		\nosp2

	moveq		#1,d2		;Set width to one to "fool" plot routine
	and.b		d5,d6		;Combine left and right masks into the right mask...
	move.b		#0,d5		;...and clear the old left mask
	sub.w		#1,a2		;Move *plot...
	subq.w		#1,a0		;...and *gfx. Necessary since this is a special case.
	subq.w		#1,a6		;...and *mask.

\nosp2	move.l		#$1228FFFF,\sm1		;Self-modifying code: use a PreByte
	move.l		#$1A2EFFFF,\sm11
	bra.s		\cont4
;(x/8) is on or to the right of (x1_clip/8), adjust destination start parameters
\xpos2	sub.w		d4,a2			;Move *plot
	move.l		#$123C0000,\sm1		;Self-modifying code: don't use a PreByte
	move.l		#$1A3C0000,\sm11

\cont4
	sub.w		d2,a4			;Source RowAdd
	move.w		W_WIDTH_IN_WORDS(a3),a5
	add.w		a5,a5			;Since window width is in words
	sub.w		d2,a5			;Target RowAdd

	subq.w		#1,d3			;We're using dbra.
	subq.w		#2,d2			;We're using dbra, plus one byte will be used outside the loop.

	and.w		#%0000000000000111,d0

;******************
;*                *
;* START PLOTTING *
;*                *
;******************
	move.w		d2,a1

\row	swap		d6			;Swap in left mask into d6.lw

\sm1	move.b		-1(a0),d1		;Self-modifying code: first byte to use before plotting
\sm11	move.b		-1(a6),d5		;Self-modifying code: first byte to use before plotting

	lsl.w		#8,d1		;Prepare for first gfx
	lsl.w		#8,d5		;Prepare for first mask
	move.b		(a0)+,d1	;First gfx on row
	move.b		(a6)+,d5	;First mask on row
	move.w		d1,d4		;Store gfx
	move.w		d5,d2		;Store mask
	lsr.w		d0,d4		;Shift gfx
	lsr.w		d0,d2		;Shift mask
	and.b		d6,d2		;Mask + left mask
	and.b		d2,d4		;Mask first gfx

	move.b		(a2),d7		;LCD
	not.b		d2		;Invert mask
	and.b		d2,d7		;Mask LCD
	or.b		d7,d4		;LCD + new gfx

	move.b		d4,(a2)+	;Plot

	move.w		a1,d7
	blt.s		\nxt
\col	lsl.w		#8,d1		;Prepare for next gfx
	lsl.w		#8,d5		;Prepare for next mask
	move.b		(a0)+,d1	;Next gfx
	move.b		(a6)+,d5	;Next mask

	move.w		d1,d4		;Store gfx
	move.w		d5,d2		;Store mask
	lsr.w		d0,d4		;Shift gfx
	lsr.w		d0,d2		;Shift mask
	and.b		d2,d4		;Mask gfx
	not.b		d2		;Invert mask
	and.b		(a2),d2		;Masked LCD (and x,y = and y,x)
	or.b		d2,d4		;LCD + new gfx
	move.b		d4,(a2)+	;Plot
	dbra.s		d7,\col

\nxt	lsl.w		#8,d1		;Prepare for last gfx
	lsl.w		#8,d5		;Prepare for last mask
\sm2	move.b		(a0),d1		;Self-modifying code: last byte to plot
\sm21	move.b		(a6),d5		;Self-modifying code: last byte to plot
	lsr.w		d0,d1		;Shift gfx
	lsr.w		d0,d5		;Shift mask

	swap		d6			;Swap in right mask into d6.lw
	and.b		d6,d5		;Mask + right mask

	and.b		d5,d1		;Mask last gfx

	move.b		(a2),d7		;LCD
	not.b		d5		;Invert gfx
	and.b		d5,d7		;Mask LCD
	or.b		d7,d1		;LCD + new gfx
	move.b		d1,(a2)		;Plot

	add.w		a4,a0		;Next row on source
	add.w		a4,a6		;Next row on source
	add.w		a5,a2		;Next row on plot
	dbra.s		d3,\row

\exit	movem.l		(sp)+,a0-a6/d0-d7
	rts


;***************************************************************************


;clopen	dc.w	1,5
;	dc.b	%11111110
;	dc.b	%11010110
;	dc.b	%11101110
;	dc.b	%11010110
;	dc.b	%11111110
;clclose	dc.w	1,5
;	dc.b	%10000010
;	dc.b	%10101010
;	dc.b	%10010010
;	dc.b	%10101010
;	dc.b	%10000010
;clmask	dc.b	%11111110
;	dc.b	%11111110
;	dc.b	%11111110
;	dc.b	%11111110
;	dc.b	%11111110

words	dc.w	%1000000000000000
	dc.w	%0100000000000000
	dc.w	%0010000000000000
	dc.w	%0001000000000000
	dc.w	%0000100000000000
	dc.w	%0000010000000000
	dc.w	%0000001000000000
	dc.w	%0000000100000000
	dc.w	%0000000010000000
	dc.w	%0000000001000000
	dc.w	%0000000000100000
	dc.w	%0000000000010000
	dc.w	%0000000000001000
	dc.w	%0000000000000100
	dc.w	%0000000000000010
	dc.w	%0000000000000001

masks	dc.b	%11111111	;Lefthand masks
	dc.b	%01111111
	dc.b	%00111111
	dc.b	%00011111
	dc.b	%00001111
	dc.b	%00000111
	dc.b	%00000011
	dc.b	%00000001

	dc.b	%10000000	;Righthand masks
	dc.b	%11000000
	dc.b	%11100000
	dc.b	%11110000
	dc.b	%11111000
	dc.b	%11111100
	dc.b	%11111110
	dc.b	%11111111

	dc.b	%10000000	;Righthand masks
	dc.b	%11000000
	dc.b	%11100000
	dc.b	%11110000
	dc.b	%11111000
	dc.b	%11111100
	dc.b	%11111110
	dc.b	%00000000

wmasks	dc.w	%1111111111111111	;Lefthand masks
	dc.w	%0111111111111111
	dc.w	%0011111111111111
	dc.w	%0001111111111111
	dc.w	%0000111111111111
	dc.w	%0000011111111111
	dc.w	%0000001111111111
	dc.w	%0000000111111111
	dc.w	%0000000011111111
	dc.w	%0000000001111111
	dc.w	%0000000000111111
	dc.w	%0000000000011111
	dc.w	%0000000000001111
	dc.w	%0000000000000111
	dc.w	%0000000000000011
	dc.w	%0000000000000001

	dc.w	%1000000000000000	;Righthand masks
	dc.w	%1100000000000000
	dc.w	%1110000000000000
	dc.w	%1111000000000000
	dc.w	%1111100000000000
	dc.w	%1111110000000000
	dc.w	%1111111000000000
	dc.w	%1111111100000000
	dc.w	%1111111110000000
	dc.w	%1111111111000000
	dc.w	%1111111111100000
	dc.w	%1111111111110000
	dc.w	%1111111111111000
	dc.w	%1111111111111100
	dc.w	%1111111111111110
	dc.w	%1111111111111111
